local Boat = ...

--------------------------------------------------------------------------
local HARBOUR_COMP_BOAT = {
	TypeName = HarbourPrefix .. "_COMP_BOAT",
	ParentType = "COMPONENT",
	Properties = {
--		{ Name = "BoatOffsetY", Type = "float", Default = 2.11 },
		{ Name = "FlagTerrain", Type = "integer", Default = 2 ^ OBJECT_FLAG.TERRAIN:toNumber() + 2 ^ OBJECT_FLAG.BUILDING_PART:toNumber() + 2 ^ OBJECT_FLAG.PLATFORM:toNumber()},
                { Name = "BoatSpeed", Type = "float", Default = 1.0 },
                { Name = "BoatLength", Type = "float", Default = 10.0 },
                { Name = "RandomPath", Type = "boolean", Default = false },
                { Name = "MaxRandomDistanceMove", Type = "float", Default = 50.0 },
                { Name = "MaxBoatNameFlagCompNumber", Type = "integer", Default = 0 },
                { Name = "_onEnabledNumber", Type = "integer", Default = 0 },
                { Name = "FirstBoatInitialPosition", Type = "boolean", Default = true },
                { Name = "BoatInitialPositionReturn", Type = "integer", Default = 0 },
                { Name = "BoatInitialPosition", Type = "vec3f" },
                { Name = "BoatInitialOrientation", Type = "quaternion" },
                { Name = "NoWater", Type = "boolean", Default = false },
                { Name = "IsSailorAssigned", Type = "boolean", Default = false },
                { Name = "IsInitDone", Type = "boolean", Default = false },
	}
}

--------------------------------------------------------------------------
-- Called by the Behavior_tree
function HARBOUR_COMP_BOAT:registerAssignedSailor(_sailor)
    Boat:log("registerAssignedSailor")
    self._assignedSailor = _sailor
    self.IsSailorAssigned = true
end
--------------------------------------------------------------------------
-- Called by the Behavior_tree
function HARBOUR_COMP_BOAT:checkBoatSailing()
        Boat:log("checkBoatSailing")
        if self.BoatInitialPositionReturn == 3 then
            Boat:log("checkBoatSailing return FALSE")
            self.BoatInitialPositionReturn = 4
            return BEHAVIOR_TREE_NODE_RESULT.FALSE
        end
        Boat:log("checkBoatSailing return TRUE")
        return BEHAVIOR_TREE_NODE_RESULT.TRUE
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:create()
    self._boatSequenceNumberPath = 0
    self._assignedSailor = nil
    self._onEnabledNumber = 0
    self.MaxBoatNameFlagCompNumber = 2 -- don't currently works with 0
    self._initTimer = 10
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:init()
        Boat:log("init: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor))
        Boat:log("init: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
        self._onEnabledNumber = 0
        self._initTimer = 10
        self.goBackToHarbourCondition = 10
        self.BoatInitialPositionReturn = 0 -- idle state
        if not self.IsSailorAssigned then
            self._assignedSailor = nil
        end
        self._boatSequenceNumberPath = 0
        self._timer = -1
        self.globalPositionPath = {}
        self.globalPositionPathIndex = 0
        self.IsInitDone = true
end


--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:onDisabled()
        self._onEnabledNumber = self._onEnabledNumber - 1
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:isSailorEmbarked()
        if self._assignedSailor == nil then
            Boat:logWarning("isSailorEmbarked: Sailor not assigned by Behavior_tree yet")
            Boat:log("isSailorEmbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber)
            Boat:log("isSailorEmbarked: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
            return false
        end
        if self._onEnabledNumber < 1 then
            Boat:logWarning("isSailorEmbarked: Boat not enabled yet")
            Boat:log("isSailorEmbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber)
            Boat:log("isSailorEmbarked: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
 
            return false
        end
        local _guardingLocationComp = self:getOwner():findFirstObjectWithComponentDown(HarbourPrefix .. "_COMP_GUARDING_LOCATION")
        if _guardingLocationComp == nil then       
            Boat:log("isSailorEmbarked: " .. HarbourPrefix .. "_COMP_GUARDING_LOCATION nil")
            return false
        end
        local deltaX = 	_guardingLocationComp:getOwner():getGlobalPosition()[1] - self._assignedSailor:getGlobalPosition()[1] 
        local deltaZ = 	_guardingLocationComp:getOwner():getGlobalPosition()[3] - self._assignedSailor:getGlobalPosition()[3] 
        local delta = deltaX*deltaX + deltaZ*deltaZ 
        if delta > 1 then
            Boat:log("isSailorEmbarked: Sailor not in Boat yet")
            Boat:log("isSailorEmbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber .. " delta = " .. delta)
            Boat:log("isSailorEmbarked: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
            return false
        end
--        Boat:log("isSailorEmbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. " delta = " .. delta)
--        Boat:log("isSailorEmbarked: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
        return true
end
--------------------------------------------------------------------------
-- currently not used
--[[
function HARBOUR_COMP_BOAT:isSailorDisbarked()
        if self._assignedSailor == nil then
            Boat:log("isSailorDisbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber)
            Boat:log("isSailorDisbarked: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
            return true -- by default the sailor is disbarked
        end
        local _guardingLocationComp = self:getOwner():findFirstObjectWithComponentDown(HarbourPrefix .. "_COMP_GUARDING_LOCATION")
        if _guardingLocationComp == nil then       
            Boat:log("isSailorDisbarked: " .. HarbourPrefix .. "_COMP_GUARDING_LOCATION nil")
            return false -- case of error
        end
        local deltaX = 	_guardingLocationComp:getOwner():getGlobalPosition()[1] - self._assignedSailor:getGlobalPosition()[1] 
        local deltaZ = 	_guardingLocationComp:getOwner():getGlobalPosition()[3] - self._assignedSailor:getGlobalPosition()[3] 
        local delta = deltaX*deltaX + deltaZ*deltaZ 
        if delta < 1 then
            Boat:log("isSailorDisbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. self._onEnabledNumber .. " delta = " .. delta)
            Boat:log("isSailorDisbarked: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
            return false
        end
        Boat:log("isSailorDisbarked: delta = " .. tostring(delta))
        Boat:log("isSailorDisbarked: RandomPath = " .. tostring(self.RandomPath) .. tostring(self._assignedSailor) .. " delta = " .. delta)
        Boat:log("isSailorDisbarked: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
        return true
end
--]]
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:launchShip()
        Boat:log("launchShip: FlagTerrain = " .. self.FlagTerrain .. " bit TERRAIN = " .. OBJECT_FLAG.TERRAIN:toNumber() .. " self._assignedSailor = " .. tostring(self._assignedSailor))
        -- memorize the position of the boat in the harbour
        if self.FirstBoatInitialPosition then
                self.BoatInitialPosition = self:getOwner():getGlobalPosition()
                self.BoatInitialOrientation = self:getOwner():getGlobalOrientation()
                self.FirstBoatInitialPosition = false
                self.globalPositionPath = {}
                Boat:log("Memorize Boat Initial Position = " .. tostring(self.BoatInitialPosition) .. " list = " .. tostring(self.globalPositionPath))
        end
        -- memorize the elevation of the worker above the boat
        self.sailorDeltaY = self:getOwner():getGlobalPosition()[2] - self._assignedSailor:getGlobalPosition()[2]
        -- Try of a simple attach of sailor to the boat, but it doesn't work
        --        if self._assignedSailor ~= nil then
        --           self._assignedSailor:setParent(self:getOwner())
        --        end
        Boat:log("End of launchShip")
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:onEnabled()
        self._onEnabledNumber = self._onEnabledNumber + 1
-- the synchronisation is now done in the update
--        if self:isSailorEmbarked() then
--            self:launchShip()
--        end
end
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:chooseRandomSequence()
        self._timer = math.random() * self.MaxRandomDistanceMove / self.BoatSpeed
        self._rotationY = (math.random() - 0.5) * 2 * math.pi
--        Boat:log("chooseRandomSequence: _rotationY = " .. tostring(self._rotationY))
        self._boatSequenceNumberPath = 2
end

--------------------------------------------------------------------------
-- for predefined path: not implemented yet
--[[
function HARBOUR_COMP_BOAT:updateBoatNextPosition(_initialSequence)
        Boat:log("updateBoatNextPosition: _initialSequence = " .. _initialSequence)
        local BoatNameFlagComp = HarbourPrefix .. "_COMP_FLAG"
        if _initialSequence < 3 then
            BoatNameFlagComp = HarbourPrefix .. "_COMP_FLAG1"
        else
            local BoatFlagNumber = _initialSequence - 1
            if BoatFlagNumber > self.MaxBoatNameFlagCompNumber then
                Boat:logWarning("No more path defined")
                return false
            end
            BoatNameFlagComp = HarbourPrefix .. "_COMP_FLAG" .. BoatFlagNumber 
        end
        Boat:log("BoatNameFlagComp = " .. BoatNameFlagComp)

        Boat:log("GlobalPosition = " .. tostring(self:getOwner():getGlobalPosition()))

        local _nextFlag = self:getLevel():find(BoatNameFlagComp, false)        
        if _nextFlag == nil then
            Boat:logWarning("No more path defined")
            return false
        end
        Boat:log("_nextFlag = " .. tostring(_nextFlag))
        if not _nextFlag.BoatIsFlagSet then
            Boat:logWarning("Flag not set")
            return false
        end
        if _nextFlag:getOwner() == nil then    
            Boat:logWarning("No flag owner defined")
            return false
        end
        Boat:log("_nextFlag:getOwner() = " .. tostring(_nextFlag:getOwner()))
        local _flagObject = _nextFlag:getOwner()
        Boat:log("_positionFlag = " .. tostring(_flagObject:getGlobalPosition()))
        self._nextPosition = _flagObject:getGlobalPosition()       
        Boat:log("_nextPosition = " .. tostring(self._nextPosition))
        return true
end

function HARBOUR_COMP_BOAT:updateBoatSequenceNumberPath(_initialSequence, _isNewPosition)
        Boat:log("_initialSequence = " .. _initialSequence .. " _isNewPosition = " .. tostring(_isNewPosition))
        if _isNewPosition then
            if _initialSequence < 3 then
                Boat:log("Step 3: move to next position")
                return 3
            else 
                Boat:log("Step  greater than 3: move to next position")
                return _initialSequence + 1
            end
        else
            if _initialSequence < 3 then

                Boat:log("Step 1: nothing to do")
                return 1 -- nothing to do
            else
                self._nextPosition = self._initialPosition
                Boat:log("Step 2: return to initial position")
                return 2 -- return to initial position
            end
        end
end
--]]
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:getSinusFromQuaternion(_q)
    return 2 * _q[4] * _q[2]
end

function HARBOUR_COMP_BOAT:getCosinusFromQuaternion(_q)
    return _q[4] * _q[4] - _q[2] * _q[2]
end

function HARBOUR_COMP_BOAT:getAngleFromQuaternion(_q)
        local sinp = self:getSinusFromQuaternion(_q)
        local cosp = self:getCosinusFromQuaternion(_q)
        local normep = cosp*cosp + sinp*sinp
        if (normep < 0.99) or (normep > 1.01)  then
            Boat:logError("getAngleFromQuaternion: getGlobalOrientation() error: " .. tostring(_q))
            return nil
        end
        local _angleCos = math.acos(cosp)
        if sinp < 0 then
            _angleCos = - _angleCos
        end
        if _angleCos < 0 then
           _angleCos = _angleCos + 2*math.pi
        end
        local _angleSin = math.asin(sinp)
        if cosp < 0 then
            _angleSin = math.pi - _angleSin
        end
        if _angleSin < 0 then
           _angleSin = _angleSin + 2*math.pi
        end
        if (_angleCos*_angleCos - _angleSin*_angleSin) > 0.01 then
            Boat:logError("getAngleFromQuaternion: updateBoatNextMove accuracy error")
        end
--        Boat:log("getAngleFromQuaternion : sinp = " .. sinp .. " cosp = " .. cosp .. " asinp = " .. (_angleSin/math.pi) .. " * pi" .. " acosp = " .. (_angleCos/math.pi) .. " * pi" .. " return = " .. (((_angleCos + _angleSin) / 2)/math.pi) .. " * pi")
        return (_angleCos + _angleSin) / 2
end

function HARBOUR_COMP_BOAT:getTimerMove(_nextPosition)
    Boat:log("getTimerMove _nextPosition = " .. tostring(_nextPosition))
    local DeltaX = _nextPosition[1] - self:getOwner():getGlobalPosition()[1]
    local DeltaZ = _nextPosition[3] - self:getOwner():getGlobalPosition()[3]
    local Distance = (DeltaX*DeltaX + DeltaZ*DeltaZ) ^ 0.5
    Boat:log("getTimerMove : X = " .. DeltaX .. " Z = " .. DeltaZ .. " Distance = " .. Distance)
    return Distance / self.BoatSpeed
end

function HARBOUR_COMP_BOAT:getRotationYFromTarget(_nextPosition, _initialBoatOrientation)
    Boat:log("getRotationYFromTarget _nextPosition = " .. tostring(_nextPosition))
    self:getOwner():globalLookAt(_nextPosition)
    local _angle = self:getAngleFromQuaternion(self:getOwner():getGlobalOrientation()) - self:getAngleFromQuaternion(_initialBoatOrientation)
    Boat:log("getRotationYFromTarget : return = " .. (_angle/math.pi) .. " * pi ; current Angle : return = " .. (self:getAngleFromQuaternion(self:getOwner():getGlobalOrientation())/math.pi) .. " * pi")
    return _angle
end

--------------------------------------------------------------------------
-- for predefined path: not implemented yet
--[[
function HARBOUR_COMP_BOAT:chooseNextFlag()
        local IsUpdatedBoatNextPosition = false
        Boat:log("IsUpdatedBoatNextPosition before = " .. tostring(IsUpdatedBoatNextPosition))
        IsUpdatedBoatNextPosition = self:updateBoatNextPosition(self._boatSequenceNumberPath)
        Boat:log("IsUpdatedBoatNextPosition after = " .. tostring(IsUpdatedBoatNextPosition))
        self._boatSequenceNumberPath = self:updateBoatSequenceNumberPath(self._boatSequenceNumberPath, IsUpdatedBoatNextPosition)
        Boat:log("Sequence after = " .. self._boatSequenceNumberPath)

        if self._boatSequenceNumberPath > 1 then
            self._rotationY = self:getRotationYFromTarget(self._nextPosition, self.BoatInitialOrientation)
            self._timer = self:getTimerMove(self._nextPosition)
        end
end
--]]
--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:checkWaterElevation(_boatDestination)
    local raycastResult = {}
--    Boat:log("checkWaterElevation: initial raycastResult = " .. tostring(raycastResult))
     
    -- Raycast from the building position + Y = 100
    -- to building position - y = 100
    -- only on objects with a WATER flag
    local FromPosition = { _boatDestination[1], _boatDestination[2]+100, _boatDestination[3] }
    local ToPosition = { _boatDestination[1], _boatDestination[2]-100, _boatDestination[3] }
--    Boat:log("checkWaterElevation: FromPosition = " .. tostring(FromPosition) .. "ToPosition = " .. tostring(ToPosition))
    
--    Boat:log("checkWaterElevation: final raycastResult = RigidBody = " .. tostring(raycastResult["RigidBody"]) .. " Position = " .. tostring(raycastResult["Position"]) .. " Normal = " .. tostring(raycastResult["Normal"]))
    if not self:getLevel():rayCast(FromPosition, ToPosition, raycastResult, 2 ^ OBJECT_FLAG.WATER:toNumber()) then
         Boat:log("FromPosition = " .. tostring(FromPosition) .. "ToPosition = " .. tostring(ToPosition))
         self.NoWater = true
         Boat:logWarning("HARBOUR_COMP_BOAT: Boat NOT on water (out of map ?)")
         return 0
    else
--        Boat:log("checkWaterElevation = " .. - _boatDestination[2] + raycastResult["Position"][2])
         return -_boatDestination[2] + raycastResult["Position"][2]
    end
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:refreshBoatRotation(_rotationY)
    self:getOwner():setRotationY(_rotationY + math.pi/2) -- +pi/2 to take into account the rotation due to the boat model
-- to be enhanced: the rotation should be smoothed and the sailor should follow the move if the GuardLocation is not close to the pivot of the ship
    if self:isSailorEmbarked() then
        self._assignedSailor:setGlobalPosition({self:getOwner():getGlobalPosition()[1], self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3]}) --security to avoid the "drifts"
-- doesn't work !!!        self._assignedSailor:setRotationY(_rotationY + math.pi/2)
    end
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:moveBoatOnDt(_dt)
    -- Raycast from the last current position to the next position + BoatLength
    -- only on objects with a TERRAIN flag
        local raycastResult = {}
        local q = self:getOwner():getGlobalOrientation()
        local sinp = self:getSinusFromQuaternion(q)
        local cosp = self:getCosinusFromQuaternion(q)
        -- take into account the orientation of the boat in the model
        local BoatExtremity = { self:getOwner():getGlobalPosition()[1] - cosp * self.BoatLength, 
                                  self:getOwner():getGlobalPosition()[2],
                                  self:getOwner():getGlobalPosition()[3] + sinp * self.BoatLength}
        if self:getLevel():rayCast(self:getOwner():getGlobalPosition(), BoatExtremity, raycastResult, self.FlagTerrain) then
            Boat:log("collision with TERRAIN raycastResult = RigidBody = " .. tostring(raycastResult["RigidBody"]) .. " Position = " .. tostring(raycastResult["Position"]) .. " Normal = " .. tostring(raycastResult["Normal"]))
            return false -- TERRAIN Reached
        end
    -- compute the next water elevation: p is the rotation since the creation of the boat. The orientation of the model is taken into account    
        local BoatDestination = { self:getOwner():getGlobalPosition()[1] - cosp * _dt * self.BoatSpeed, 
                                     self:getOwner():getGlobalPosition()[2],
                                     self:getOwner():getGlobalPosition()[3] + sinp * _dt * self.BoatSpeed}
        local BoatDeltaY = self:checkWaterElevation(BoatDestination)
        if self.NoWater then
            return false
        end
    -- translations 
        self:getOwner():translate({ math.sin(self._rotationY) * _dt * self.BoatSpeed, BoatDeltaY, math.cos(self._rotationY) * _dt * self.BoatSpeed })
        if self._assignedSailor ~= nil then
            local q = self._assignedSailor:getGlobalOrientation()
            local sins = self:getSinusFromQuaternion(q)
            local coss = self:getCosinusFromQuaternion(q)
            self._assignedSailor:translate({ (coss*math.sin(self._rotationY) + sins*math.cos(self._rotationY)) * _dt * self.BoatSpeed, 
                                             BoatDeltaY - self.sailorDeltaY,
                                             (coss*math.cos(self._rotationY) - sins*math.sin(self._rotationY)) * _dt * self.BoatSpeed })
        end
        return true -- ok    
end

--------------------------------------------------------------------------
function HARBOUR_COMP_BOAT:update()
--        Boat:log("update: RandomPath = " .. tostring(self.RandomPath))
--        Boat:log("update: _boatSequenceNumberPath = " .. self._boatSequenceNumberPath)
    local dt = self:getLevel():getDeltaTime()

-- Sequences for the boat returning to harbour
    if self.BoatInitialPositionReturn > 0 then
        if self.BoatInitialPositionReturn == 1 then
            Boat:log("goBackToHarbourCondition (status 1) self:getOwner():getGlobalPosition = " .. tostring(self:getOwner():getGlobalPosition()) .. " self.globalPositionPathIndex = " .. self.globalPositionPathIndex)
            if self.globalPositionPathIndex <= 0 then
-- for unitary tests: infinite loop
--                self.globalPositionPathIndex = #self.globalPositionPath
                Boat:log("Restore Boat Initial Position (status 1) = " .. tostring(self.BoatInitialPosition))
                self:getOwner():setGlobalTransform(self.BoatInitialPosition, self.BoatInitialOrientation)
                if self._assignedSailor ~= nil then
                    self._assignedSailor:setGlobalPosition(self.BoatInitialPosition)
                end
                self.BoatInitialPositionReturn = 3 -- warn the behavior_tree
                return
            end 
--[[
            if self.globalPositionPathIndex < #self.globalPositionPath then -- workaround of an error which I didn't find the cause    
                self:getOwner():setGlobalTransform(self.globalPositionPath[self.globalPositionPathIndex+1], self.BoatInitialOrientation)
                if self._assignedSailor ~= nil then
                    self._assignedSailor:setGlobalPosition(self.globalPositionPath[self.globalPositionPathIndex+1])
                end
            end
--]]
            Boat:log("goBackToHarbourCondition (status 1) list = " .. tostring(self.globalPositionPath) .. " list[" .. self.globalPositionPathIndex .. "] = " .. tostring(self.globalPositionPath[self.globalPositionPathIndex]))
            self._initTimer = self:getTimerMove(self.globalPositionPath[self.globalPositionPathIndex])
            if self._initTimer < 1 then -- self.BoatLength / self.BoatSpeed then
                Boat:log("goBackToHarbourCondition same position")
                self.globalPositionPathIndex = self.globalPositionPathIndex - 1   
                return
            end 
            self._rotationY = self:getRotationYFromTarget(self.globalPositionPath[self.globalPositionPathIndex], self.BoatInitialOrientation )
            self:refreshBoatRotation(self._rotationY)
            self.globalPositionPathIndex = self.globalPositionPathIndex - 1
            Boat:log("goBackToHarbourCondition (status 1): self._rotationY = " .. self._rotationY .. " self._initTimer = " .. self._initTimer)
            self.BoatInitialPositionReturn = 2 -- start moving
        elseif self.BoatInitialPositionReturn == 2 then
            Boat:log("goBackToHarbourCondition (status 2): self._rotationY = " .. self._rotationY .. " self._initTimer = " .. self._initTimer .. " self.globalPositionPathIndex = " .. self.globalPositionPathIndex)
            if self._initTimer >= 0 then
                self._initTimer = self._initTimer - dt
-- for unitary tests: only rotations, no moving
--                Boat:log("Move back to Position: self._initTimer = " .. self._initTimer)
                if self:moveBoatOnDt(dt) then
                    return
                end
                Boat:log("Move back to Position: Terrain ?")
            end
            self.BoatInitialPositionReturn = 1 -- new back path
            return
        elseif self.BoatInitialPositionReturn == 3 then
            -- seems update is not recalled while the workplace is disabled
            -- so I can't check the disbarkment, and "jump" to the embarkment
        elseif self.BoatInitialPositionReturn == 4 then
            Boat:log("Restore Boat Initial Position status = 4")
            self.BoatInitialPositionReturn = 0 -- idle state
            self._initTimer = 10 -- wait 
            self._boatSequenceNumberPath = 0 -- wait sailor embarked
            return
        end

-- sequences for launching boat (first boat setting after a sailor assignment, after a return to harbour, or after a reloaded game)
    elseif self._boatSequenceNumberPath == 0 then
        -- case of reload
        if self._initTimer >= 0 then
            self._initTimer = self._initTimer - dt
            return
        end
        Boat:log("Launch boat wait for init done")
        if not self.IsInitDone then
            self._initTimer = 10
            return
        end
        Boat:log("Launch boat _boatSequenceNumberPath = 0  _onEnabledNumber = 0" .. self._onEnabledNumber)
        if self._onEnabledNumber == 0 then
            self._initTimer = 10
            return
        end
        Boat:log("Launch boat _boatSequenceNumberPath = 0, _onEnabledNumber > 0")
        if not self:isSailorEmbarked() then
            self._initTimer = 10
            return
        end
        Boat:log("Launch boat _boatSequenceNumberPath = 0, sailor embarked")
        self:launchShip()
        self._boatSequenceNumberPath = 1 -- initialization casual moving
        self.globalPositionPath = {} -- empty the list of returning path
        self.goBackToHarbourCondition = 10 -- no immediate return to harbour
        self._timer = -1 -- to first set the orientation of the boat

-- sequences for casual moves
    -- orientation setting
    elseif self._timer < 0 then
        if self.RandomPath then
            -- count down to the next return to harbour        
            self.goBackToHarbourCondition = self.goBackToHarbourCondition - 1
            if self.goBackToHarbourCondition <= 0 then
-- for unitary tests
--[[
            self.globalPositionPath = {
                { self:getOwner():getGlobalPosition()[1]-100, self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3]-100 },
                { self:getOwner():getGlobalPosition()[1]-100, self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3]},
                { self:getOwner():getGlobalPosition()[1]-100, self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3]+100 },
                { self:getOwner():getGlobalPosition()[1], self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3]+100 },
                { self:getOwner():getGlobalPosition()[1]+100, self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3]+100},
                { self:getOwner():getGlobalPosition()[1]+100, self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3] },
                { self:getOwner():getGlobalPosition()[1]+100, self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3]-100 },
                { self:getOwner():getGlobalPosition()[1], self:getOwner():getGlobalPosition()[2], self:getOwner():getGlobalPosition()[3]-100 }
            }
--]]
                self.globalPositionPathIndex = #self.globalPositionPath
--                self.globalPositionPathIndex = 1 -- force the preceeding behavior: direct return back to the initial position
                self.BoatInitialPositionReturn = 1 -- start return back
                return                            
            end
            -- memorize the last position
            table.insert(self.globalPositionPath, self:getOwner():getGlobalPosition())
            Boat:log("memorize path list = " .. tostring(self.globalPositionPath) .. " list[" .. #self.globalPositionPath .. "] = " .. tostring(self.globalPositionPath[#self.globalPositionPath]))
            Boat:log("memorize path list = " .. tostring(self.globalPositionPath))
            -- choose a random destination
            self:chooseRandomSequence()
        else
            -- choose a predefined path (Todo: the question is that the flags cannot be set outside the area of the harbour monument)
--          self:chooseNextFlag()
        end
        -- rotation          
        if self._boatSequenceNumberPath > 1 then
            self:refreshBoatRotation(self._rotationY)
        end

    -- straight moving
    else      
        self._timer = self._timer - dt 
        if self._boatSequenceNumberPath > 1 then
            if not self:moveBoatOnDt(dt) then
                self._timer = -1 -- Force the end of the loop
            end 
        end
    end

-- error cases
    if self.NoWater then --error detected by checkWaterElevation
        self:getOwner():destroy()
    end
end

--------------------------------------------------------------------------
Boat:registerClass(HARBOUR_COMP_BOAT)
